import numpy as np
from typing import Callable, Any


def fgh_basis(
    n: int,
    x: np.array,
    N: int,
    A: float,
    B: float,
):
    L = B - A
    if N % 2 == 0:
        l_idx = np.arange(-N // 2, N // 2)  # l ranges from -N/2 to N/2-1
    else:
        l_idx = np.arange(-(N - 1) // 2, (N - 1) // 2 + 1)  # l ranges from -(N-1)/2 to (N-1)/2

    # print(l_idx.shape[0])
    exponent = 2j * np.pi * l_idx[:, None] * ((x[None, :] - A) / L - n / N)
    summation = np.sum(np.exp(exponent), axis=0)  # sum each column
    varphi = summation / np.sqrt(N * L)

    return varphi


def fgh_hamiltonian(
    N: int,
    x: np.ndarray,
    V: Callable[..., Any],
    is_code_optimization: bool = True,
):
    # Initialize the Hamiltonian matrix
    ham = np.zeros((N, N), dtype=np.float64)

    # Calculate grid spacing and wave vector
    dx = x[1] - x[0]
    K = 2.0 * np.pi / dx

    if is_code_optimization:
        # Precompute indices differences and sine/cosine values
        indices = np.arange(N)
        i_minus_j = indices[:, None] - indices[None, :]
        sin_term = np.sin(np.pi * i_minus_j / N)

        # Handle diagonal elements separately
        diagonal_factor = (N**2 + 2) / 12.0 if N % 2 == 0 else (N**2 - 1) / 12.0
        diagonal = 0.5 * diagonal_factor * (K / N) ** 2 + V(x)

        # Compute off-diagonal elements using broadcasting
        with np.errstate(divide="ignore", invalid="ignore"):
            # print(i_minus_j)
            off_diagonal = 0.5 * (-1) ** np.abs(i_minus_j)
            if N % 2 == 0:
                off_diagonal /= sin_term**2
            else:
                cos_term = np.cos(np.pi * i_minus_j / N)
                off_diagonal *= cos_term / (sin_term**2)

        # Eliminate infinities or NaNs caused by division by zero on diagonal
        np.fill_diagonal(off_diagonal, 0)

        # Scale the off-diagonal terms
        off_diagonal *= 0.5 * (K / N) ** 2

        # Combine diagonal and off-diagonal components
        np.fill_diagonal(ham, diagonal)
        ham += off_diagonal

    else:
        # Legacy code (not optimized, 20~50 times slower than optimized code)
        if N % 2 == 0:
            for i in range(N):
                for j in range(N):
                    if j == i:
                        ham[i, j] = 0.5 * ((N**2 + 2) / 12.0) * (K / N) ** 2 + V(x[i])
                    else:
                        ham[i, j] = 0.5 * (0.5 * (-1) ** (i - j) / ((np.sin(np.pi * (i - j) / N)) ** 2)) * (K / N) ** 2
        else:
            for i in range(N):
                for j in range(N):
                    if j == i:
                        ham[i, j] = 0.5 * ((N**2 - 1) / 12.0) * (K / N) ** 2 + V(x[i])
                    else:
                        ham[i, j] = 0.5 * (0.5 * (-1) ** (i - j) * np.cos(np.pi * (i - j) / N) / ((np.sin(np.pi * (i - j) / N)) ** 2)) * (K / N) ** 2

    return ham
