import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as tck


# plt.rcParams["axes.prop_cycle"] = plt.cycler(
#    color=["#000000", "#ff0000", "#0000ff", "#228b22", "#ff00ff", "#ffd900", "#00ffff", "#7f00ff", "#ff7f00", "#00ff7f", "#bebebe", "#666666"]
# )


def plot_xy(
    x_dict: dict[str, np.ndarray],
    y_dict: dict[str, np.ndarray],
    y_dict_error: dict[str, np.ndarray] = {},
    output_image_name: str | None = None,  # "out.png"
    show_gui: bool = True,
    data_optional: dict[str, dict] | None = None,
    title_name: str = "x-y graph",
    title_font_size: int | str = "default",
    title_v_offset: float | int | str = "default",
    figure_size: tuple[int, int] = (7, 5),
    aspect_ratio: str | float = "auto",
    legend_loc: tuple[float, float] | str | None = None,  # ,"upper left",
    legend_transparent: float = 0.5,
    legend_bg_color: str = "default",
    legend_edge_color: str = "default",
    x_scale: str = "linear",
    x_base: float | int | str = "auto",
    x_start_tick: float = None,
    x_end_tick: float = None,
    x_range: tuple[float, float] | str = "auto",
    x_major_ticks_manual: tuple | None = None,
    x_label: str = "$x$",
    x_grid: str = "major",
    x_ticks_direction: str = "in",
    is_x_minor_ticks: bool = False,
    x_minor_ticks_manual: tuple | None = None,
    x_minor_division: int = 1,
    x_invert: bool = False,
    y_scale: str = "linear",
    y_base: float | int | str = "auto",
    y_range: tuple[float, float] | str = "auto",
    y_start_tick: float = None,
    y_end_tick: float = None,
    y_major_ticks_manual: tuple | None = None,
    y_label: str = "$y$",
    y_grid: str = "major",
    y_ticks_direction: str = "in",
    is_y_minor_ticks: bool = False,
    y_minor_ticks_manual: tuple | None = None,
    y_minor_division: int = 1,
    y_invert: bool = False,
    default_font_family: str = "Times New Roman",
    default_font_size: int = 10,
    default_font_set: str = "stix",
) -> None:
    plt.rcdefaults()

    plt.rcParams["font.family"] = default_font_family  # font
    plt.rcParams["font.size"] = default_font_size  # character size
    plt.rcParams["mathtext.fontset"] = default_font_set  # font Math

    # ===================
    # 1. Global settings
    # ===================

    # - Make figure and subplots
    fig, axs = plt.subplots(1, 1, figsize=figure_size)

    # - Aspect ratio
    axs.set_aspect(aspect_ratio, adjustable="box")

    # - Plot data and it's label in legend
    # print(data_optional, show_error_bar)
    if data_optional is None:
        for k in y_dict.keys():
            # if show_error_bar:
            if y_dict_error.get(k) is not None:
                axs.errorbar(x_dict[k], y_dict[k], yerr=y_dict_error[k], elinewidth=1, fmt=".-", ms=3, label=k, capsize=2, capthick=1)
            else:
                axs.plot(x_dict[k], y_dict[k], ".-", ms=3, label=k)
    else:
        for k in y_dict.keys():
            p = data_optional.get(k)
            if p is not None:
                fmt = p.get("fmt") if p.get("fmt") is not None else ".-"
                data_label = p.get("ti") if p.get("ti") is not None else k
                line_width = p.get("lw") if p.get("lw") is not None else 1
                line_color = p.get("lc") if p.get("lc") is not None else "C0"
                marker_size = p.get("ms") if p.get("ms") is not None else 3
                e_line_width = p.get("ew") if p.get("ew") is not None else line_width / 2
                cap_size = p.get("cs") if p.get("cs") is not None else 2
                cap_thick = p.get("ct") if p.get("ct") is not None else 1
                if y_dict_error.get(k) is not None:
                    axs.errorbar(
                        x_dict[k],
                        y_dict[k],
                        yerr=y_dict_error[k],
                        fmt=fmt,
                        lw=line_width,
                        ms=marker_size,
                        label=data_label,
                        color=line_color,
                        elinewidth=e_line_width,
                        capsize=cap_size,
                        capthick=cap_thick,
                    )
                else:
                    axs.plot(x_dict[k], y_dict[k], fmt, lw=line_width, ms=marker_size, label=data_label, color=line_color)
            else:
                if y_dict_error.get(k) is not None:
                    axs.errorbar(
                        x_dict[k],
                        y_dict[k],
                        yerr=y_dict_error[k],
                        label=k,
                        elinewidth=1,
                        capsize=2,
                        capthick=1,
                    )
                else:
                    axs.plot(
                        x_dict[k],
                        y_dict[k],
                        label=k,
                    )

    # - Legend
    legend_location = legend_loc if isinstance(legend_loc, (tuple, str)) else plt.rcParams["legend.loc"]
    framealpha = legend_transparent if isinstance(legend_transparent, float) else plt.rcParams["legend.framealpha"]
    legend_facecolor = legend_bg_color if legend_bg_color != "default" else None
    legend_edgecolor = legend_edge_color if legend_edge_color != "default" else None
    axs.legend(
        loc=legend_location,
        framealpha=framealpha,
        fancybox=True,
        facecolor=legend_facecolor,
        edgecolor=legend_edgecolor,
    )

    # - Title
    i0 = title_font_size if isinstance(title_font_size, int) else plt.rcParams["font.size"]
    f0 = title_v_offset if isinstance(title_v_offset, (int, float)) else plt.rcParams["axes.titlepad"]
    axs.set_title(title_name, pad=f0, fontsize=i0)

    # ========
    # 2. Axis
    # ========
    axs.grid(False)
    axs.minorticks_off()
    # -----------
    # 2.1 x-axis
    # -----------

    # - x Label
    axs.set_xlabel(x_label)

    # - x Grid
    if x_grid in ["major", "both"]:
        axs.grid(axis="x", which="major", color="#cccccc", linestyle="-")  # show x-axis grid
    if x_grid in ["minor", "both"]:
        axs.grid(axis="x", which="minor", color="#e7e7e7", linestyle="--")  # show x-axis sub-grid

    # - x Ticks
    axs.tick_params(axis="x", which="both", direction=x_ticks_direction)
    if x_scale == "linear":
        axs.set_xscale(x_scale)
        # -- Major ticks location
        if x_major_ticks_manual is None:
            if x_start_tick is None and x_end_tick is None:
                if isinstance(x_base, (float, int)):
                    axs.xaxis.set_major_locator(tck.MultipleLocator(base=float(x_base)))
                else:
                    axs.xaxis.set_major_locator(tck.MaxNLocator(nbins="auto"))
            else:
                if isinstance(x_base, (float, int)):
                    xstart, xend = axs.get_xlim()
                    if x_start_tick is not None:
                        xstart = x_start_tick if x_range != "auto" else x_range[0]
                    if x_end_tick is not None:
                        xend = x_end_tick if x_range != "auto" else x_range[1]
                    ticks = np.arange(xstart, xend, x_base)
                    axs.set_xticks(ticks)
                else:
                    axs.xaxis.set_major_locator(tck.MaxNLocator(nbins="auto"))

            # -- Major ticks label format
            mj_formatter = tck.ScalarFormatter()
            mj_formatter.set_scientific(True)  # Enable Scientific notation
            mj_formatter.set_powerlimits(plt.rcParams["axes.formatter.limits"])
            mj_formatter.set_useMathText(True)  # (False)1e8, (True)10^8
            axs.xaxis.set_major_formatter(mj_formatter)
        else:
            axs.set_xticks(x_major_ticks_manual[0])
            axs.xaxis.set_major_formatter(tck.FixedFormatter(x_major_ticks_manual[1]))

        # -- Minor ticks location
        if is_x_minor_ticks:
            axs.xaxis.set_minor_formatter(tck.NullFormatter())
            if x_minor_ticks_manual is None:
                axs.xaxis.set_minor_locator(tck.AutoMinorLocator(n=x_minor_division))
            else:
                axs.xaxis.set_minor_locator(tck.FixedLocator(x_minor_ticks_manual))

    elif x_scale == "log":
        axs.set_xscale(x_scale)
        x_log_base = x_base if isinstance(x_base, (float, int)) else 10.0

        if x_major_ticks_manual is None:
            # -- Major ticks location
            # Calc the position at integer power
            a_min_list = []
            a_max_list = []
            for k in x_dict.keys():
                a_min_list.append(np.amin(x_dict[k]))
                a_max_list.append(np.amax(x_dict[k]))
            a_min = np.amin(a_min_list)
            a_max = np.amin(a_max_list)
            n_power_min = np.floor(np.log(a_min) / np.log(x_log_base)).astype(int)
            n_power_max = np.ceil(np.log(a_max) / np.log(x_log_base)).astype(int)
            ticks = [x_log_base**i for i in range(n_power_min - 1, n_power_max + 2)]
            # Set the ticks location clearly
            axs.xaxis.set_major_locator(tck.FixedLocator(ticks))
            # -- Major ticks label format
            mj_formatter = tck.LogFormatterMathtext(base=x_log_base, labelOnlyBase=True)
            axs.xaxis.set_major_formatter(mj_formatter)
        else:
            axs.xaxis.set_major_locator(tck.FixedLocator(x_major_ticks_manual[0]))
            axs.xaxis.set_major_formatter(tck.FixedFormatter(x_major_ticks_manual[1]))

        # -- Minor ticks location
        if is_x_minor_ticks:
            axs.xaxis.set_minor_formatter(tck.NullFormatter())
            if x_minor_division is None:
                axs.xaxis.set_minor_locator(tck.LogLocator(base=x_log_base, subs=None))
            else:
                axs.xaxis.set_minor_locator(tck.LogLocator(base=x_log_base, subs=np.arange(1.0, x_minor_division, 1) * (1.0 / x_minor_division)))
    else:
        raise ValueError(f'x_scale must be "linear" or "log", your x_scale={x_scale}')

    # - x Range
    if isinstance(x_range, tuple):
        axs.set_xlim(x_range[0], x_range[1])
    else:
        axs.set_xlim(auto=True)

    if x_invert:
        axs.invert_xaxis()

    # -----------
    # 2.2 y-axis
    # -----------

    # - y Label
    axs.set_ylabel(y_label)

    # - y Grid
    if y_grid in ["major", "both"]:
        axs.grid(axis="y", which="major", color="#cccccc", linestyle="-")  # show y-axis grid
    if y_grid in ["minor", "both"]:
        axs.grid(axis="y", which="minor", color="#e7e7e7", linestyle="--")  # show y-axis sub-grid

    # - y Ticks
    axs.tick_params(axis="y", which="both", direction=y_ticks_direction)
    if y_scale == "linear":
        axs.set_yscale(y_scale)

        if y_major_ticks_manual is None:
            if y_start_tick is None and y_end_tick is None:
                if isinstance(y_base, (float, int)):
                    axs.yaxis.set_major_locator(tck.MultipleLocator(base=float(y_base)))
                else:
                    axs.yaxis.set_major_locator(tck.MaxNLocator(nbins="auto"))
            else:
                if isinstance(y_base, (float, int)):
                    # define ticks by start and interval
                    ystart, yend = axs.get_ylim()
                    if y_start_tick is not None:
                        ystart = y_start_tick if y_range != "auto" else y_range[0]
                    if y_end_tick is not None:
                        yend = y_end_tick if y_range != "auto" else y_range[1]
                    ticks = np.arange(ystart, yend, y_base)
                    axs.set_yticks(ticks)
                else:
                    axs.yaxis.set_major_locator(tck.MaxNLocator(nbins="auto"))
            # -- Major ticks label format
            mj_formatter = tck.ScalarFormatter()
            mj_formatter.set_scientific(True)  # Enable Scientific notation
            mj_formatter.set_powerlimits(plt.rcParams["axes.formatter.limits"])
            #   (set_powerlimits: Normal notation range, otherwise Scientific notation.)
            mj_formatter.set_useMathText(True)  # (False)1e8, (True)10^8
            axs.yaxis.set_major_formatter(mj_formatter)
        else:
            # axs.set_yticks(y_major_ticks_manual[0])
            axs.yaxis.set_major_locator(tck.FixedLocator(y_major_ticks_manual[0]))
            axs.yaxis.set_major_formatter(tck.FixedFormatter(y_major_ticks_manual[1]))

        # -- Minor ticks location
        if is_y_minor_ticks:
            axs.yaxis.set_minor_formatter(tck.NullFormatter())
            if y_minor_ticks_manual is None:
                axs.yaxis.set_minor_locator(tck.AutoMinorLocator(n=y_minor_division))
            else:
                axs.yaxis.set_minor_locator(tck.FixedLocator(y_minor_ticks_manual))

    elif y_scale == "log":
        axs.set_yscale(y_scale)
        y_log_base = y_base if isinstance(y_base, (float, int)) else 10.0

        if y_major_ticks_manual is None:
            # -- Major ticks location
            # Calc the position at integer power
            a_min_list = []
            a_max_list = []
            for k in y_dict.keys():
                a_min_list.append(np.amin(y_dict[k]))
                a_max_list.append(np.amax(y_dict[k]))
            a_min = np.amin(a_min_list)
            a_max = np.amin(a_max_list)
            if a_min <= 1e-100:
                a_min = 1e-100
            n_power_min = np.floor(np.log(a_min) / np.log(y_log_base)).astype(int)
            n_power_max = np.ceil(np.log(a_max) / np.log(y_log_base)).astype(int)
            ticks = [y_log_base**i for i in range(n_power_min - 1, n_power_max + 2)]
            # Set the ticks location clearly
            axs.yaxis.set_major_locator(tck.FixedLocator(ticks))
            # -- Major ticks label format
            mj_formatter = tck.LogFormatterMathtext(base=y_log_base, labelOnlyBase=True)
            axs.yaxis.set_major_formatter(mj_formatter)
        else:
            axs.set_yticks(y_major_ticks_manual[0])
            axs.yaxis.set_major_formatter(tck.FixedFormatter(y_major_ticks_manual[1]))

        # -- Minor ticks location
        if is_y_minor_ticks:
            axs.yaxis.set_minor_formatter(tck.NullFormatter())
            if y_minor_division is None:
                axs.yaxis.set_minor_locator(tck.LogLocator(base=y_log_base, subs=None))
            else:
                axs.yaxis.set_minor_locator(tck.LogLocator(base=y_log_base, subs=np.arange(1.0, y_minor_division, 1) * (1.0 / y_minor_division)))
    else:
        raise ValueError(f'y_scale must be "linear" or "log", your y_scale={y_scale}')

    # - y Range
    if isinstance(y_range, tuple):
        axs.set_ylim(y_range[0], y_range[1])
    else:
        axs.set_ylim(auto=True)

    if y_invert:
        axs.invert_yaxis()
    # ==========
    # 3. Output
    # ==========
    # - Pop up GUI
    if show_gui:
        plt.show()

    # - output image
    if output_image_name is not None:
        if output_image_name.endswith(".eps"):
            # Don't plot the left-bottom region because the Bounding Box has negative value
            fig.subplots_adjust(left=0.3, right=0.9, bottom=0.3, top=0.9)
        # Save figure
        fig.savefig(output_image_name, dpi=400, bbox_inches="tight")
    plt.close()
